/*
 * Copyright (c) 2014-2019,2021,2023 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *					Anthony Mallet on Wed Feb 19 2014
 */
#include "acoptitrack.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <assert.h>
#include <err.h>
#include <errno.h>
#include <float.h>
#include <fnmatch.h>
#include <math.h>
#include <netdb.h>
#include <poll.h>
#include <stddef.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "natnet.h"

#define MAX_NAMELENGTH 256
#define MAX_PACKETSIZE 0x10000	/* actual packet size is dynamic */

#define NAT_PING			0
#define NAT_PINGRESPONSE		1
#define NAT_REQUEST			2
#define NAT_RESPONSE			3
#define NAT_REQUEST_MODELDEF		4
#define NAT_MODELDEF			5
#define NAT_REQUEST_FRAMEOFDATA		6
#define NAT_FRAMEOFDATA			7
#define NAT_MESSAGESTRING		8
#define NAT_UNRECOGNIZED_REQUEST	100
#define NAT_UNDEFINED			999999.9999

#define NAT_DESCR_MSET			0
#define NAT_DESCR_BODY			1
#define NAT_DESCR_SKEL			2
#define NAT_DESCR_CAMPOS		5

struct natnet_sender {
  char name[MAX_NAMELENGTH];		/* sending app's name */
  unsigned char version[4];		/* app's version [X.X.X.X] */
  unsigned char natnet_version[4];	/* NatNet version [X.X.X.X] */
  uint64_t highresclockfrequency;	/* only with natnet >= 3.0 */
} __attribute__((__packed__));

struct natnet_packet {
  uint16_t id;		/* message ID (e.g. NAT_FRAMEOFDATA) */
  uint16_t bytes;	/* actual payload size */
  union {
    char	c[MAX_PACKETSIZE];
    struct natnet_sender sender;
  } data;
};

struct natnet_context {
  int cmdfd, datafd;
  struct sockaddr_in cmdaddr, dataaddr;

  struct natnet_sender sender;
  struct natnet_descr descr;

  struct natnet_packet msg;
  struct natnet_frame frame;

  double acq_ts, offset_ts;
  struct timeval origin_ts;
};

static int	natnet_recv(natnet_context_s ctx, int fd, size_t *sz,
                            char *emsg, size_t emsgsize);
static int	natnet_cmd_socket(natnet_context_s ctx, uint16_t local_port,
                                  const char *host, const char *host_port,
                                  char *emsg, size_t emsgsize);
static int	natnet_data_socket(natnet_context_s ctx,
                                   const char *mcast, const char *mcast_port,
                                   char *emsg, size_t emsgsize);
static uint32_t	natnet_version(natnet_context_s ctx);
static size_t	strnfilter(char *dst, size_t n, const char *src,
                           const char *accept, char repl);
static struct timeval
		get_ts(uint32_t ds, double period, double *ats, double *offset,
                       time_t origin_ts);


/* --- natnet_ping --------------------------------------------------------- */

/** Send/recv a ping command to the host
 */
int
natnet_snd_ping(natnet_context_s ctx, char *emsg, size_t emsgsize)
{
  struct natnet_packet msg;
  size_t len;
  int e;
  assert(ctx->cmdfd >= 0);

  msg.id = NAT_PING;
  msg.bytes = 0;
  len = offsetof(struct natnet_packet, data);

  e = sendto(ctx->cmdfd, &msg, len, 0,
             (struct sockaddr *)&ctx->cmdaddr, sizeof(ctx->cmdaddr));
  if (e < 0) {
    snprintf(emsg, emsgsize, "send: %s", strerror(errno));
    return -1;
  }
  return 0;
}

int
natnet_rcv_ping(natnet_context_s ctx, char *info, size_t infosize)
{
  assert(ctx && ctx->msg.id == NAT_PINGRESPONSE);

  snprintf(info, infosize, "%s-%d.%d.%d.%d, natnet-%d.%d.%d.%d",
           ctx->msg.data.sender.name,
           ctx->msg.data.sender.version[0],
           ctx->msg.data.sender.version[1],
           ctx->msg.data.sender.version[2],
           ctx->msg.data.sender.version[3],
           ctx->msg.data.sender.natnet_version[0],
           ctx->msg.data.sender.natnet_version[1],
           ctx->msg.data.sender.natnet_version[2],
           ctx->msg.data.sender.natnet_version[3]);
  return 0;
}


/* --- natnet_modeldef ----------------------------------------------------- */

/** Send/recv a model definition request to the host
 */
int
natnet_snd_modeldef(natnet_context_s ctx, char *emsg, size_t emsgsize)
{
  struct natnet_packet msg;
  size_t len;
  int e;
  assert(ctx->cmdfd >= 0);

  msg.id = NAT_REQUEST_MODELDEF;
  msg.bytes = 0;
  len = offsetof(struct natnet_packet, data);

  e = sendto(ctx->cmdfd, &msg, len, 0,
             (struct sockaddr *)&ctx->cmdaddr, sizeof(ctx->cmdaddr));
  if (e < 0) {
    snprintf(emsg, emsgsize, "send: %s", strerror(errno));
    return -1;
  }
  return 0;
}

const struct natnet_descr *
natnet_rcv_modeldef(natnet_context_s ctx)
{
  union { const char *c; const uint32_t *u; const float *f; const void *v; } p;
  struct natnet_msetdescr *mset;
  struct natnet_bodydescr *body;
  uint32_t k, n, m, i, j;

  assert(ctx && ctx->msg.id == NAT_MODELDEF);

  p.c = ctx->msg.data.c;
  ctx->descr.nmsets = 0;
  ctx->descr.nbodies = 0;

  n = *p.u++;
  for(i = 0; i < n; i++) {
    k = *p.u++;
    switch(k) {
      case NAT_DESCR_MSET:
        while (ctx->descr.nmsets >= ctx->descr.maxmsets) {
          mset = realloc(ctx->descr.mset,
                         (1+ctx->descr.maxmsets) * sizeof(*mset));
          if (!mset) return NULL;
          ctx->descr.mset = mset;
          ctx->descr.mset[ctx->descr.maxmsets].nmnames = 0;
          ctx->descr.mset[ctx->descr.maxmsets].mname = NULL;
          ctx->descr.maxmsets++;
        }
        mset = &ctx->descr.mset[ctx->descr.nmsets++];
        if (mset->mname) {
          for (j = 0; j < mset->nmnames; j++)
            free(mset->mname[j]);
          free(mset->mname);
        }

        mset->name = strdup(p.c);
        if (!mset->name) return NULL;
        strnfilter(mset->safename, sizeof(mset->safename),
                   mset->name, "[a-zA-Z0-9_]", '_');
        p.c = strchr(p.c, 0)+1;
        mset->nmnames = *p.u++;
        mset->mname = calloc(mset->nmnames, sizeof(*mset->mname));
        if (!mset->mname) return NULL;
        for (j = 0; j < mset->nmnames; j++) {
          mset->mname[j] = strdup(p.c);
          if (!mset->mname[j]) return NULL;
          p.c = strchr(p.c, 0)+1;
        }
        break;

      case NAT_DESCR_BODY:
        while (ctx->descr.nbodies >= ctx->descr.maxbodies) {
          body = realloc(ctx->descr.body,
                         (1+ctx->descr.maxbodies) * sizeof(*body));
          if (!body) return NULL;
          ctx->descr.body = body;
          ctx->descr.body[ctx->descr.maxbodies].name = NULL;
          ctx->descr.maxbodies++;
        }
        body = &ctx->descr.body[ctx->descr.nbodies++];
        if (body->name) free(body->name);

        body->name = strdup(p.c);
        if (!body->name) return NULL;
        strnfilter(body->safename, sizeof(body->safename),
                   body->name, "[a-zA-Z0-9_]", '_');
        p.c = strchr(p.c, 0)+1;
        body->id = *p.u++;
        body->parent = *p.u++;
        body->offset.x = *p.f++;
        body->offset.y = *p.f++;
        body->offset.z = *p.f++;

        if (natnet_version(ctx) >= 3000) {
          m = *p.u++; /* n markers */
          p.f += 3 * m; /* pos */
          p.u += m; /* active labels */
        }
        break;

      case NAT_DESCR_SKEL: /* ignore */
        p.c = strchr(p.c, 0)+1; /* name */
        p.u++; /* id */
        m = *p.u++;
        for(j=0; j<m; j++) {
          p.c = strchr(p.c, 0)+1; /* name */
          p.u++; /* id */
          p.u++; /* parent */
          p.f++; /* x */
          p.f++; /* y */
          p.f++; /* z */
        }
        break;

      case NAT_DESCR_CAMPOS: /* ignore */
        p.c = strchr(p.c, 0)+1; /* name */
        p.f += 3; /* x, y, z */
        p.f += 4; /* qx, qy, qz, qw */
        break;

      default:
        warnx("unknown data description %d", k);
        assert(!"unknown data description");
        break;
    }
  }

  /* end */
  assert(p.c == ctx->msg.data.c + ctx->msg.bytes);

  return &ctx->descr;
}


/* --- natnet_frameofdata -------------------------------------------------- */

/** Send/recv a frame of data
 */
const struct natnet_frame *
natnet_rcv_frameofdata(natnet_context_s ctx)
{
  uint32_t previd = ctx->frame.id;

  union {
    const char *c;
    const uint16_t *s;
    const uint32_t *u;
    const uint64_t *l;
    const float *f;
    const double *d;
    const void *v;
  } p;
  struct natnet_mset *mset;
  struct natnet_body *body;
  uint32_t n, m, o, i, j;

  assert(ctx && ctx->msg.id == NAT_FRAMEOFDATA);
  p.c = ctx->msg.data.c;
  ctx->frame.id = *p.u++;

  /* marker sets */
  if (ctx->frame.maxmsets < *p.u) {
    mset = realloc(ctx->frame.mset, *p.u * sizeof(*mset));
    if (!mset) return NULL;
    ctx->frame.mset = mset;

    mset = &ctx->frame.mset[ctx->frame.maxmsets];
    for(j = ctx->frame.maxmsets; j < *p.u; j++, mset++) {
      mset->mset.n = mset->mset.max = 0;
      mset->mset.pos = NULL;
    }
    ctx->frame.maxmsets = *p.u;
  }
  ctx->frame.nmsets = *p.u++;

  mset = ctx->frame.mset;
  for(i = 0; i < ctx->frame.nmsets; i++, mset++) {
    mset->name = p.c;
    p.c = strchr(mset->name, 0)+1;
    if (mset->mset.max < *p.u) {
      void *pos = realloc(mset->mset.pos, *p.u * sizeof(*mset->mset.pos));
      if (!pos) return NULL;
      mset->mset.pos = pos;
      mset->mset.max = *p.u;
    }
    mset->mset.n = *p.u++;
    for(j = 0; j < mset->mset.n; j++) {
      mset->mset.pos[j].x = *p.f++;
      mset->mset.pos[j].y = *p.f++;
      mset->mset.pos[j].z = *p.f++;
    }

    mset->descr = NULL;
    for (j = 0; j < ctx->descr.nmsets; j++)
      if (!strcmp(ctx->descr.mset[j].name, mset->name)) {
        mset->descr = &ctx->descr.mset[j];
        break;
      }
  }

  /* unindentified markers - deprecated but still present */
  n = *p.u++;
  p.f += 3 * n;

  /* rigid bodies */
  if (ctx->frame.maxbodies < *p.u) {
    body = realloc(ctx->frame.body, *p.u * sizeof(*body));
    if (!body) return NULL;
    ctx->frame.body = body;
    body = &ctx->frame.body[ctx->frame.maxbodies];
    for(j = ctx->frame.maxbodies; j < *p.u; j++, body++) {
      body->mset.n = body->mset.max = 0;
      body->mset.pos = NULL;
      body->mid = NULL;
      body->msize = NULL;
    }
    ctx->frame.maxbodies = *p.u;
  }
  ctx->frame.nbodies = *p.u++;
  body = ctx->frame.body;
  for(i = 0; i < ctx->frame.nbodies; i++, body++) {
    body->id = *p.u++;
    body->pos = p.v;
    p.v = &body->pos[1];

    /* natnet before 3.0 */
    if (natnet_version(ctx) < 3000) {
      if (body->mset.max < *p.u) {
        void *pos = realloc(body->mset.pos, *p.u * sizeof(*body->mset.pos));
        if (!pos) return NULL;
        body->mset.pos = pos;
        body->mset.max = *p.u;
      }
      body->mset.n = *p.u++;
      for (j = 0; j < body->mset.n; j++) {
        body->mset.pos[j].x = *p.f++;
        body->mset.pos[j].y = *p.f++;
        body->mset.pos[j].z = *p.f++;
      }

      body->mid = p.v;
      p.v = &body->mid[body->mset.n];
      body->msize = p.v;
      p.v = &body->msize[body->mset.n];
    } else
      body->mset.n = 0;

    body->error = *p.f++;

    /* natnet 2.6 and later */
    if (natnet_version(ctx) >= 2006) {
      uint16_t params = *p.s++;
      if (!(params & 1)) /* body not tracked */ body->pos = NULL;
    }

    body->descr = NULL;
    for (j = 0; j < ctx->descr.nbodies; j++)
      if (ctx->descr.body[j].id == body->id) {
        body->descr = &ctx->descr.body[j];
        break;
      }
  }

  /* skeletons - ignore */
  n = *p.u++;
  for(i = 0; i < n; i++) {
    p.u++; /* id */
    m = *p.u++; /* n bodies */
    for(j = 0; j < m; j++) {
      p.u++; /* id */
      p.f += 7; /* pos */

      /* natnet 3.0 and later dropped this */
      if (natnet_version(ctx) < 3000) {
        o = *p.u++; /* n markers */
        p.f += 3 * o; /* pos */
        p.u += o; /* ids */
        p.f += o; /* sizes */
      }

      p.f++; /* error */
    }
  }

  /* labeled and unlabeled markers */
  ctx->frame.aset.n = 0;
  n = *p.u++;
  for(i = 0; i < n; i++) {
    o = *p.u++; /* id == (model << 16) + marker id */

    if (natnet_version(ctx) >= 3000) {
      uint16_t model = o >> 16;
      uint16_t id = o & 0xffff;

      switch(model) {
        case 0:
          if (ctx->frame.aset.max <= ctx->frame.aset.n) {
            void *pos =
              realloc(ctx->frame.aset.pos,
                      (ctx->frame.aset.max+1)*sizeof(*ctx->frame.aset.pos));
            if (!pos) return NULL;
            ctx->frame.aset.pos = pos;
            ctx->frame.aset.max++;
          }
          ctx->frame.aset.pos[ctx->frame.aset.n].x = p.f[0];
          ctx->frame.aset.pos[ctx->frame.aset.n].y = p.f[1];
          ctx->frame.aset.pos[ctx->frame.aset.n].z = p.f[2];
          ctx->frame.aset.n++;
          break;

        default:
          if (id < 1 || id > 255) break;

          body = ctx->frame.body;
          for(j = 0; j < ctx->frame.nbodies; j++, body++) {
            if (body->id != model) continue;
            if (body->mset.max <= body->mset.n) {
              void *pos =
                realloc(body->mset.pos,
                        (body->mset.max+1)*sizeof(*body->mset.pos));
              if (!pos) return NULL;
              body->mset.pos = pos;
              body->mset.max++;
            }
            body->mset.pos[body->mset.n].x = p.f[0];
            body->mset.pos[body->mset.n].y = p.f[1];
            body->mset.pos[body->mset.n].z = p.f[2];
            body->mset.n++;
            break;
          }
          break;
      }
    }
    p.f += 3; /* pos */
    p.f++; /* size */

    /* natnet 2.6 and later */
    if (natnet_version(ctx) >= 2006) {
      p.s++; /* marker params (occluded, solved point cloud/model) */
    }

    /* natnet 3.0 and later */
    if (natnet_version(ctx) >= 3000) {
      p.f++; /* residual */
    }
  }

  /* force plate data - ignore */
  if (natnet_version(ctx) >= 2009) {
    n = *p.u++;
    for(i = 0; i < n; i++) {
      p.u++; /* id */
      m = *p.u++; /* n channels */
      for(j = 0; j < m; j++) {
        o = *p.u++; /* n frames */
        p.f += o; /* values */
      }
    }
  }

  /* device data - ignore */
  if (natnet_version(ctx) >= 2011) {
    n = *p.u++;
    for(i = 0; i < n; i++) {
      p.u++; /* id */
      m = *p.u++; /* n channels */
      for(j = 0; j < m; j++) {
        o = *p.u++; /* n frames */
        p.f += o; /* values */
      }
    }
  }

  /* latency */
  if (natnet_version(ctx) < 3000)
    ctx->frame.latency = *p.f++;

  /* timecode */
  ctx->frame.tcode = *p.u++;
  ctx->frame.tsub = *p.u++;

  /* timestamp - this seems to be the frame.id divided by the current frame
   * rate.
   * natnet 2.7 uses double precision
   * natnet 2.6 uses single precision
   * natnet 2.5 does not have any */
  if (natnet_version(ctx) < 2006) {
    gettimeofday(&ctx->frame.timestamp, NULL);
    ctx->acq_ts = 0.;
  } else {
    double ts;
    if (natnet_version(ctx) >= 2007) {
#ifdef __ARM_EABI__
      /* doubles must be aligned on 8 bytes */
      memcpy(&ts, p.v, sizeof(ts));
      p.d++;
#else
      ts = *p.d++;
#endif
    } else if (natnet_version(ctx) >= 2006)
      ts = *p.f++;
    else assert(0);

    ctx->frame.timestamp = get_ts(
      ctx->frame.id - previd, ts / ctx->frame.id,
      &ctx->acq_ts, &ctx->offset_ts, ctx->origin_ts.tv_sec);

    /* natnet 3.0 and later */
    if (natnet_version(ctx) >= 3000) {
      int64_t dt;
      double latency;

      dt = - *p.l++; /* camera mid-exposure ts */
      p.l++; /* camera received ts */
      dt += *p.l++; /* transmit ts */

      latency = (double)dt / ctx->sender.highresclockfrequency;
      ctx->frame.timestamp.tv_sec -= floor(latency);
      ctx->frame.timestamp.tv_usec -= (latency - floor(latency)) * 1e6;
      if (ctx->frame.timestamp.tv_usec < 0) {
        ctx->frame.timestamp.tv_usec += 1000000;
        ctx->frame.timestamp.tv_sec--;
      }
    }
  }

  /* frame params - natnet 2.6 and later */
  if (natnet_version(ctx) >= 2006) {
    uint16_t params = *p.s++;
    ctx->frame.refresh = !!(params & 2);
  } else
    ctx->frame.refresh = 0;

  /* end */
  p.u++;
  assert(p.c == ctx->msg.data.c + ctx->msg.bytes);

  return &ctx->frame;
}


/* --- natnet_poll --------------------------------------------------------- */

/** Wait for an event on sockets
 */
int
natnet_poll(natnet_context_s ctx, struct timeval *timeout)
{
  assert(ctx && (ctx->cmdfd >= 0 || ctx->datafd >= 0));

  struct pollfd fds[2] = {
    { .fd = ctx->cmdfd, .events = POLLIN },
    { .fd = ctx->datafd, .events = POLLIN }
  };
  int e;

  do {
    e = poll(fds, 2, timeout->tv_sec * 1000 + timeout->tv_usec / 1000);
  } while(e < 0 && errno == EINTR);
  if (e < 0) {
    /* other poll failures are not supposed to happen here */
    assert(!"unexpected poll() error");
    return NATNET_EVERR;
  }

  if (e > 0) {
    e = 0;
    if (fds[0].revents) e |= NATNET_EVCMD;
    if (fds[1].revents) e |= NATNET_EVDAT;
  }
  return e;
}


/* --- natnet_recv --------------------------------------------------------- */

/** Read a message on the 'command' or 'data' socket
 */
int
natnet_cmd_recv(natnet_context_s ctx, size_t *sz,
                char *emsg, size_t emsgsize)
{
  return natnet_recv(ctx, ctx->cmdfd, sz, emsg, emsgsize);
}

int
natnet_data_recv(natnet_context_s ctx, size_t *sz,
                 char *emsg, size_t emsgsize)
{
  return natnet_recv(ctx, ctx->datafd, sz, emsg, emsgsize);
}

static int
natnet_recv(natnet_context_s ctx, int fd, size_t *sz,
            char *emsg, size_t emsgsize)
{
  struct sockaddr_in src;
  socklen_t len = sizeof(src);
  ssize_t s;
  assert(fd >= 0);

  s = recvfrom(fd, &ctx->msg, sizeof(ctx->msg), 0,
               (struct sockaddr *)&src, &len);
  if (s <= 0) {
    if (s < 0)
      snprintf(emsg, emsgsize, "recv: %s", strerror(errno));
    if (s == 0)
      snprintf(emsg, emsgsize, "recv: connection reset by peer");
    return NATNET_MSGERR;
  }
  if (sz) *sz = s;

  /* filter source address */
  if (fd == ctx->cmdfd) {
    if (len != sizeof(src)) return 0;
    if (src.sin_addr.s_addr != ctx->cmdaddr.sin_addr.s_addr) return 0;
    if (src.sin_port != ctx->cmdaddr.sin_port &&
        src.sin_port != ctx->dataaddr.sin_port) return 0;
  } else if (fd == ctx->datafd) {
    if (len != sizeof(src)) return 0;
    if (src.sin_addr.s_addr != ctx->cmdaddr.sin_addr.s_addr) return 0;
    if (src.sin_port != ctx->dataaddr.sin_port) return 0;
  } else assert(0);

  switch (ctx->msg.id) {
    case NAT_PINGRESPONSE: return NATNET_MSGPING;
    case NAT_MODELDEF: return NATNET_MSGMODELDEF;
    case NAT_FRAMEOFDATA: return NATNET_MSGFRAMEOFDATA;

    case NAT_UNRECOGNIZED_REQUEST: return 0;
  }

  assert(!"unknown message id");
  return NATNET_MSGNONE;
}


/* --- natnet_connect ------------------------------------------------------ */

natnet_context_s
natnet_connect(uint16_t local_port, const char *host, const char *host_port,
               const char *mcast, const char *mcast_port,
               char *emsg, size_t emsgsize)
{
  natnet_context_s ctx;
  int done;
  int e, s;

  ctx = calloc(1, sizeof(*ctx));
  if (!ctx) {
    snprintf(emsg, emsgsize, "malloc: %s", strerror(errno));
    return NULL;
  }
  ctx->cmdfd = ctx->datafd = -1;
  ctx->offset_ts = -DBL_MAX;
  gettimeofday(&ctx->origin_ts, NULL);

  /* 'command' socket */
  if (natnet_cmd_socket(ctx, local_port, host, host_port, emsg, emsgsize)) {
    natnet_disconnect(ctx);
    return NULL;
  }

  /* 'data' socket(s) */
  if (natnet_data_socket(ctx, mcast, mcast_port, emsg, emsgsize)) {
    natnet_disconnect(ctx);
    return NULL;
  }

  /* ping server (to get natnet version) */
  for(done = 3; done; --done) {
    if (natnet_snd_ping(ctx, emsg, emsgsize))
      break;

    e = natnet_poll(ctx, &(struct timeval){ .tv_sec = 1, .tv_usec = 0 });
    if (e == NATNET_EVERR) {
      snprintf(emsg, emsgsize, "unknown error");
      break;
    }

    if (e & NATNET_EVDAT) {
      s = natnet_data_recv(ctx, NULL, emsg, emsgsize);
      if (s == NATNET_MSGERR)
        break;
    }
    if (e & NATNET_EVCMD) {
      s = natnet_cmd_recv(ctx, NULL, emsg, emsgsize);
      if (s == NATNET_MSGERR)
        break;
      if (s == NATNET_MSGPING) {
        ctx->sender = ctx->msg.data.sender;
        break;
      }
    }
  }
  if (!ctx->sender.name[0] || natnet_version(ctx) < 2003 /* need >=2.3 */) {
    snprintf(emsg, emsgsize, "transport endpoint is not connected");
    natnet_disconnect(ctx);
    return NULL;
  }

  return ctx;
}


/* --- natnet_cmd_socket --------------------------------------------------- */

/** Create the 'command' socket sending commands to the host
 */
static int
natnet_cmd_socket(natnet_context_s ctx, uint16_t local_port,
                  const char *host, const char *host_port,
                  char *emsg, size_t emsgsize)
{
  struct addrinfo hints, *res, *res0;
  struct sockaddr_in addr;
  const char *cause;
  int fd;
  int e;
  assert(ctx->cmdfd < 0);

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = AF_INET; /* motive supports only IPv4 */
  hints.ai_socktype = SOCK_DGRAM;
  e = getaddrinfo(host, host_port, &hints, &res0);
  if (e) {
    snprintf(emsg, emsgsize, "%s", gai_strerror(e));
    return -1;
  }

  for (res = res0; res; res = res->ai_next) {
    fd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (fd < 0) { cause = "socket"; goto log; }

    /* bind the socket locally on a fixed port number. This is only to
     * workaround an issue with unicast streaming. Since it is not possible to
     * 'disconnect' a client, at least we guarantee that all connections to the
     * same host will not end up in multiple streams. */
    addr.sin_family = AF_INET;
    addr.sin_port = htons(local_port);
    addr.sin_addr.s_addr = INADDR_ANY;
    if (bind(fd, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
      cause = "bind to local port";
      close(fd);
      goto log;
    }

    /* send an empty packet to detect local errors */
    if (sendto(fd, NULL, 0, 0, res->ai_addr, res->ai_addrlen)) {
      cause = "send";
      close(fd);
      goto log;
    }

    ctx->cmdfd = fd;
    ctx->cmdaddr = *(struct sockaddr_in *)res->ai_addr;
    break;

log:
    {
      char loghost[64], logport[16];
      if (!getnameinfo(res->ai_addr, res->ai_addrlen,
                       loghost, sizeof(loghost),
                       logport, sizeof(logport), NI_NUMERICHOST|NI_NUMERICSERV))
        warnx("failed to reach %s:%s...", loghost, logport);
    }
  }
  freeaddrinfo(res0);

  if (ctx->cmdfd < 0) {
    snprintf(emsg, emsgsize, "%s: %s", cause, strerror(errno));
    return -1;
  }

  return 0;
}


/* --- natnet_data_socket -------------------------------------------------- */

/** Create the 'data' socket receiving multicast marker messages
 */
static int
natnet_data_socket(natnet_context_s ctx,
                   const char *mcast, const char *mcast_port,
                   char *emsg, size_t emsgsize)
{
  struct addrinfo hints, *res, *res0;
  struct ip_mreq mreq;
  const char *cause;
  int fd;
  int e;
  assert(ctx->datafd < 0);

  /* allow empty mcast address to make multicast optional */
  if (!mcast[0]) return 0;

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = AF_INET; /* motive supports only IPv4 */
  hints.ai_socktype = SOCK_DGRAM;
  hints.ai_flags = AI_PASSIVE;
  e = getaddrinfo(mcast, mcast_port, &hints, &res0);
  if (e) {
    snprintf(emsg, emsgsize, "%s", gai_strerror(e));
    return -1;
  }

  for (res = res0; res; res = res->ai_next) {
    fd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (fd < 0) { cause = "socket"; goto log; }

    if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &(int){1}, sizeof(int))) {
      cause = "setsockopt(SO_REUSEADDR)";
      close(fd);
      goto log;
    }

    memset(&mreq, 0, sizeof(mreq));
    mreq.imr_multiaddr = ((struct sockaddr_in *)res->ai_addr)->sin_addr;
    mreq.imr_interface.s_addr = INADDR_ANY;

    e = setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq));
    if (e) {
      cause = "setsockopt(IP_ADD_MEMBERSHIP)";
      close(fd);
      goto log;
    }

    if (bind(fd, res->ai_addr, res->ai_addrlen) < 0) {
      cause = "bind";
      close(fd);
      goto log;
    }

    ctx->datafd = fd;
    ctx->dataaddr = *(struct sockaddr_in *)res->ai_addr;
    break;

log:
    {
      char loghost[64], logport[16];
      if (!getnameinfo(res->ai_addr, res->ai_addrlen,
                       loghost, sizeof(loghost),
                       logport, sizeof(logport), NI_NUMERICHOST|NI_NUMERICSERV))
        warnx("failed to reach %s:%s...", loghost, logport);
    }
  }
  freeaddrinfo(res0);

  if (ctx->datafd <= 0) {
    snprintf(emsg, emsgsize, "%s: %s", cause, strerror(errno));
    return -1;
  }

  return 0;
}


/* --- natnet_disconnect --------------------------------------------------- */

void
natnet_disconnect(natnet_context_s ctx)
{
  uint32_t i, j;

  if (!ctx) return;

  if (ctx->cmdfd >= 0) close(ctx->cmdfd);
  if (ctx->datafd >= 0) close(ctx->datafd);

  if (ctx->descr.mset) {
    for(i = 0; i < ctx->descr.nmsets; i++) {
      if (ctx->descr.mset[i].name) free(ctx->descr.mset[i].name);
      if (ctx->descr.mset[i].mname) {
        for(j = 0; j < ctx->descr.mset[i].nmnames; j++)
          if (ctx->descr.mset[i].mname[j]) free(ctx->descr.mset[i].mname[j]);
        free(ctx->descr.mset[i].mname);
      }
    }
    free(ctx->descr.mset);
  }
  if (ctx->descr.body) {
    for(i = 0; i < ctx->descr.nbodies; i++) {
      if (ctx->descr.body[i].name) free(ctx->descr.body[i].name);
    }
    free(ctx->descr.body);
  }

  free(ctx);
}


/* --- natnet_version ------------------------------------------------------ */

static uint32_t
natnet_version(natnet_context_s ctx)
{
  return
    ctx->sender.natnet_version[0]*1000 +
    ctx->sender.natnet_version[1];
}


/* --- strnfilter ---------------------------------------------------------- */

static size_t
strnfilter(char *dst, size_t n, const char *src, const char *accept, char repl)
{
  char x[2];

  x[1] = '\0';
  while(--n && *src) {
    x[0] = *src++;
    if (fnmatch(accept, x, 0))
      *dst++ = repl;
    else
      *dst++ = x[0];
  }
  *dst = '\0';
  return n;
}



/* --- get_ts -------------------------------------------------------------- */

/** Implements Olson, Edwin. "A passive solution to the sensor synchronization
 * problem." International conference on Intelligent Robots and Systems (IROS),
 * 2010 IEEE/RSJ */
static struct timeval
get_ts(uint32_t ds, double period, double *ats, double *offset,
       time_t origin_ts)
{
  struct timeval tv;
  double ts;

  /* delta samples */
  if (ds > 1024) {
    /* if too many samples were lost, maybe motive was restarted.
     * reset the offset */
    *offset = -DBL_MAX;
    *ats = 0.;
  } else {
    /* consider only a super tiny (0.001%) clock drift on the sender side */
    *offset -= 1e-3/100. * ds * period;

    /* update remote timestamp */
    *ats += ds * period;
  }

  /* update offset */
  gettimeofday(&tv, NULL);
  ts = (tv.tv_sec - origin_ts) + 1e-6 * tv.tv_usec;
  if (*ats - ts > *offset)
    *offset = *ats - ts;

  /* local timestamp - reset offset if it diverged too much from realtime
   * (maybe motive didn't publish data in time or something) */
  if (ts - (*ats - *offset) > 50 * period)
    *offset = *ats - ts;
  else
    ts = *ats - *offset;

  tv.tv_sec = floor(ts);
  tv.tv_usec = (ts - tv.tv_sec) * 1e6;
  tv.tv_sec += origin_ts;
  return tv;
}
